/***************************************************************************
* BasicGrouper.C														   *
*																		   *
* Implementation of a straight-forward stroke grouping algorithm.		   *
*																		   *
*	Changes:															   *
*																		   *
*		10/20/99  arvo	  Initial coding.								   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
#include <assert.h>
#include "../util/Math.h"
#include "../rec/Font.h"
#include "../rec/BasicGrouper.h"

BasicGrouper::BasicGrouper()
{
	rec = NULL;
	lookahead = MaxLookahead - 1;
	Clear();
}

void BasicGrouper::Clear()
{
	num_strokes = 0;
	for( int i = 0; i < MaxLookahead; i++ )
		for( int j = 0; j < MaxLookahead; j++ )
		{
			prob[i][j] = 0.0;
		}
}

void BasicGrouper::operator<<( const Multiline &stroke )
{
	assert( num_strokes < MaxLookahead );
	strokes[ num_strokes++ ] = stroke;
}

// Determine how many of the strokes from the head of the queue should be
// grouped together to form a character.  Return a pointer to the symbol
// (which contains the name of the character), and the actual strokes
// grouped as a single multiline.  NULL is returned as the function value
// when the grouper does not have enough strokes to make a good decision.
// When "force" is true, NULL is returned ONLY when the queue is empty.
//
Symbol *BasicGrouper::GetCharacter( Multiline &character, int force )
{
	Symbol *symbol = NULL;
	character.Clear();
	int count = num_strokes;
	if( num_strokes == 0 ) return NULL;
	if( num_strokes < lookahead && !force ) return NULL;
	
	float min = 0.0;
	int limit = Min( num_strokes, MaxSymbolStrokes );
	for( int k = 1; k <= limit; k++ )
	{
		SearchStatus status;
		float prob = FindBestGrouping( 0, num_strokes - k, min, status );
		if( prob > min )
		{
			min   = prob;
			count = status.count;
		}
	}
	
	if( count > 0 )
	{
		// Remove the first "count" strokes from the queue.
		character = Remove( count );
		
		// Invoke the recognizer one more time to get the name of the character.
		// (It's more trouble than it's worth to retain this information from the
		// previous invokations of the recognizer.)  Use the name to look up the 
		// font information.
		Profile prof;
		rec->Recognize( character, prof );
		symbol = prof.BestSymbol();
		if( symbol != NULL )
			symbol = FontTable::FT.Lookup( symbol->Name(), "<font>" );
	}
	
	return symbol;
}

// Examine the strokes in the current queue indexed from "i" to "j" and
// determine how to best partition them into characters.  Do not consider
// any collection of strokes for which the recognizer returns a confidence
// less than "min".  Return the "status" of the search, which includes the
// number of strokes in the first and second characters, and the minimum
// confidence associated with this grouping.
//
float BasicGrouper::FindBestGrouping( int i, int j, float min, SearchStatus &status )
{
	status.Init();
	int steps = Min( lookahead, j - i + 1 );
	
	// Handle the special case of i == j.  This is where the
	// recursion bottoms out.
	if( steps <= 0 )
	{
		status.min	 = 1000.0;
		status.count = 0;
		status.next  = 0;
		return status.min;
	}
	if( steps == 1 )
	{
		status.min	 = GetProb( i, j );
		status.count = 1;
		status.next  = 0;
		return status.min;
	}
	
	for( int k = 1; k <= steps; k++ )
	{
		// Try grouping the first k strokes from the front of the queue.
		float k_prob = GetProb( i, i + k - 1 );
		
		// If the grouping is sufficiently probable, see if the rest of the
		// strokes can also be grouped with sufficiently high confidence.
		if( k_prob > min ) 
		{
			SearchStatus temp;
			FindBestGrouping( i + k, j, min, temp );
			// If all the strokes can be grouped with higher confidence
			// than the previous minimum, save the relevant information
			// about this grouping and increase the minimum.  This will
			// allow many future searches to terminate earlier.
			if( temp.min > min )
			{
				min = Min( k_prob, temp.min );
				status.min	 = min;
				status.count = k;
				status.next  = temp.count;
			}
		}
	}
	return min;
}


// Get the probability (or "confidence") that the strokes with indices 
// between "from_index" and "to_index", inclusive, form a character when 
// grouped together.  If this probability has already been computed, simply 
// retrieve the cached value.  If not, invoke the recognizer to supply the 
// value, and cache it for future reference.  This cache is what allows
// the recognizer to be invoked only O(n^2) times rather than O(2^n) times,
// where "n" is the number of strokes in the queue.  That is, while there
// are O(2^n) possible partitions of n ordered strokes, there are only O(n^2)
// contiguous sequences of strokes.
//
float BasicGrouper::GetProb( int from_index, int to_index )
{
	if( from_index > to_index ) return 0.0;
	assert( from_index < lookahead );
	assert( to_index   < lookahead );
	if( prob[ from_index ][ to_index ] == 0.0 )
	{
		Multiline S;
		for( int i = from_index; i <= to_index; i++ )
			S += strokes[i];
		Profile prof;
		rec->Recognize( S, prof );
		prob[ from_index ][ to_index ] = prof.BestProb();
		// cerr << form( "Prob (%d,%d) = %f, [%s]\n", from_index, to_index, 
		//				  prof.BestProb(), prof.BestSymbol()->Name() );
	}
	return prob[ from_index ][ to_index ];
}


// Remove the first n strokes from the queue, and shift the others
// down.  Also keep all the probabilities in the "prob" cache that 
// are still applicable; that is, all the entries (i,j) with i >= n 
// and j >= i, as these do not involve the n strokes that were just
// removed.  These entries are shifted down by (-n,-n) and zero is
// stored at the original location (to indicate a missing entry,
// which may be filled in later by GetProb).
//
Multiline BasicGrouper::Remove( int n )
{
	int i;
	Multiline character;
	for( i = 0; i < n; i++ )
	{
		character += strokes[i];
		strokes[i].Clear();
	}
	for( i = n; i < num_strokes; i++ )
	{
		strokes[ i - n ] = strokes[ i ];
		strokes[ i ].Clear();
		for( int j = i; j < num_strokes; j++ )
		{
			prob[ i - n ][ j - n ] = prob[ i ][ j ];
			prob[ i ][ j ] = 0.0;
		}
	}
	num_strokes -= n;
	return character;
}

#include "stdafx.h"
