/***************************************************************************
* Features.C															   *
*																		   *
* This module defines all of the individual "features" that are extracted  *
* from each character to identify it.  Each feature is a scalar, or a	   *
* collection of scalars, that is inserted into a "feature vector".		   *
* The feature vector is then treated as a point in "feature space". 	   *
*																		   *
*	Changes:															   *
*																		   *
*		09/09/99 arvo	  Modified many of the feature computations.	   *
*		08/15/99 arvo	  Replaced arrays with lists.					   *
*		04/04/98 arvo	  Modified feature vector & many small changes.    *
*		01/23/98 fedorov  Recoded to use Multiline. 					   *
*		01/05/97 guo	  Initial coding.								   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
#include "stdafx.h"

#include <assert.h>
#include <string.h>
#include <math.h>
#include <windows.h>
#include <fstream>
using namespace std;
#include "../util/DirectoryTree.h"
#include "../rec/Font.h"
#include "../rec/Features.h"

//extern const char *data_dir;
const char *data_dir = "data\\";
extern const char *font_dir;

/***************************************************************************
* The FeatureVector class encapsulates the vector of floats that represent
* distinct features of hand-printed characters.  The class is more convenient
* than an array, since it provides the "<<" operator for inserting features.
***************************************************************************/
void FeatureVector::Clear()
{
	for( register int i = 0; i < Max_Features; i++ ) 
	{
		feature[i] = 0;
		comment[i] = NULL;
	}
	index = 0;
}

/****************************************************************************
* Return the comment associated with feature i, if one exists.	If not, return
* the comment with the largest index that is smaller than i.  In either case,
* append the comment with the index indicating how many NULL comments exist
* between i and the comment actually returned (plus one).
****************************************************************************/
const char *FeatureVector::Comment( int i ) const
{
	static char buff[256];
	for( int j = i; j >= 0; j-- )
	{
		if( comment[j] != NULL ) 
		{
			if( j == i ) return comment[j];
			sprintf( buff, "%s: entry %d", comment[j], i - j + 1 );
			return buff;
		}
	}
	return "undefined";
}

/****************************************************************************
* The FeatureStatsVector class encapsulates the statistics that characterize
* a collection of feature vectors for the same symbol (written in the same 
* style).  The class is more convenient than an array, since it provides the 
* "<<" operator for inserting feature statistics.
****************************************************************************/
FeatureStats::FeatureStats()
{
	Clear();
}

void FeatureStats::Clear()
{
	Average =  0.0;
	StndDev =  0.0;
	Min 	=  1.0E10;
	Max 	= -1.0E10;
}

FeatureStatsVector::FeatureStatsVector()
{
	Clear();
}

void FeatureStatsVector::Clear()
{
	index = 0;
	num_samples = 0;
	for( register int i = 0; i < Max_Features; i++ ) stats[i].Clear();
}

FeatureStatsVector &FeatureStatsVector::operator<<( const FeatureStats &s )
{
	assert( index < Max_Features );
	stats[ index++ ] = s;
	return *this;
}

void FeatureStatsVector::operator+=( const FeatureVector &vec )
{
	index = vec.Size();
	for( register int i = 0; i < vec.Size(); i++ )
	{
		double x = vec.Feature(i);
		stats[i].Average += x;
		stats[i].StndDev += x * x;
		if( x > stats[i].Max ) stats[i].Max = x;
		if( x < stats[i].Min ) stats[i].Min = x;
	}
	num_samples++;
}

void FeatureStatsVector::Close()
{
	if( num_samples == 0 ) return;
	double scale = 1.0 / double( num_samples );
	for( int i = 0; i < Size(); i++ ) 
	{
		stats[i].Average *= scale;
		if( Size() == 1 )
			stats[i].StndDev = 0.0;
		else stats[i].StndDev = sqrt( stats[i].StndDev * scale - Sqr( stats[i].Average ) );
	}
}

/****************************************************************************
* Compute how many points lie in quadrants I, II, III and IV.
****************************************************************************/
static void QuadrantHistogram( const Multiline &lines, FeatureVector &F )
{
	double q1 = 0;
	double q2 = 0;
	double q3 = 0;
	double q4 = 0; 
	Vec2  Cent( lines.Center() );
	
	for( Pnode *node = lines.First(); node != NULL; node = node->Next() )
	{
		switch( Quadrant( *node - Cent ) )
		{
		case 1: q1++; break;
		case 2: q2++; break;
		case 3: q3++; break;
		case 4: q4++; break;
		}
	}
	double scale = 1.0 / lines.Points();
	F << "Fraction of points in each quadrant"
		<< q1 * scale
		<< q2 * scale
		<< q3 * scale
		<< q4 * scale;
}

/****************************************************************************
* Compute the number of points with X coordinate smaller than the average of 
* minimum and overall average (f21), between this value and overall average 
* (f22), betweem overall average and average of maximum and overall average 
* (f23), greater than this value (f24). Same values are computed for the Y 
* coordinate and stored in f25,f26,f27,f28. 	 
****************************************************************************/
void SectorHistogram( const Multiline &lines, FeatureVector &F )
{
	Box2d Box	 = lines.Box();
	Vec2  Cent	 = lines.Center();
	double LowerX = 0.5 * ( Box.Xmin() + Cent.X() );
	double UpperX = 0.5 * ( Box.Xmax() + Cent.X() );
	double LowerY = 0.5 * ( Box.Ymin() + Cent.Y() );
	double UpperY = 0.5 * ( Box.Ymax() + Cent.Y() );
	
	double x0 = 0;
	double x1 = 0;
	double x2 = 0;
	double x3 = 0;
	double y0 = 0;
	double y1 = 0;
	double y2 = 0;
	double y3 = 0;
	
	for( Pnode *node = lines.First(); node != NULL; node = node->Next() )
	{
		double x = node->X();
		double y = node->Y();
		
		if( x < LowerX	) x0++; else
			if( x < Cent.X()) x1++; else
				if( x < UpperX	) x2++; else x3++;
				
				if( y < LowerY	) y0++; else 
					if( y < Cent.Y()) y1++; else 
						if( y < UpperY	) y2++; else y3++; 
	}
	
	double scale = 1.0 / lines.Points();
	
	F << "Ratio of points in each X quarter-strip"
		<< x0 * scale
		<< x1 * scale 
		<< x2 * scale 
		<< x3 * scale;
	
	F << "Ratio of points in each Y quarter-strip"
		<< y0 * scale
		<< y1 * scale 
		<< y2 * scale 
		<< y3 * scale;
}

/****************************************************************************
* Compute several histograms and functionals based on the angles formed
* by consecutive line segments.  Normalize all values so that the rate of
* drawing has a negligible effect.
****************************************************************************/
static void AngleHistogram( const Multiline &lines, FeatureVector &F )
{
	double a0	= 0;  // Fraction of angles between  0	and  90 degrees.
	double a1	= 0;  // Fraction of angles between  90 and 180 degrees.
	double a2	= 0;  // Fraction of angles between 180 and 270 degrees.
	double a3	= 0;  // Fraction of angles between 270 and 360 degrees.
	double sum1	= 0;  // Average of positive angles.
	double sum2	= 0;  // Average of negaitve angles.
	double sum3	= 0;  // Average of squared angles.
	double sum4	= 0;  // Average of angles to 20'th power.
	double sum5	= 0;  // Average of angles to 50'th power.
	double wsum	= 0;  // Weighted sum of absolute angles.
	double scale = 1.0;
	if( lines.Points() >= 3 ) 
	{
		Pnode *prev = lines.First();
		Pnode *curr = prev->Next();
		Pnode *next = curr->Next();
		while( next != NULL )
		{
			if( !curr->Jump() && !next->Jump() )
			{
				double A_len;
				double B_len;
				Vec2  A = Unit( *curr - *prev, A_len );
				Vec2  B = Unit( *next - *curr, B_len );
				double angle = ( 1.0 + A * B ) * 0.5;
				double apow	= pow( angle, 10 );
				
				wsum += angle * ( A_len + B_len );
				
				if( Sign( A ^ B ) >= 0.0 ) 
				{
					if( angle < 0.95 ) a0++; else a1++;
					sum1 += apow;
				}
				else
				{ 
					if( angle < 0.95 ) a3++; else a2++;
					sum2 += apow;
				}
				sum3 += angle * angle;
				sum4 += apow;
				sum5 += pow( angle, 50 );
			}
			prev = curr;
			curr = next;
			next = next->Next();
		}  
		scale = 2.0 / ( a0 + a1 + a2 + a3 );
	}
	
	// Now insert the features into the feature vector.  Include
	// comment strings so that the role of these features can be
	// analyzed later.
	
	F << "Angle histogram"
		<< a0 * scale
		<< a1 * scale
		<< a2 * scale
		<< a3 * scale;
	
	F << "Weighted angle sum"		 << wsum / ( 2.0 * lines.Diam() + 1.0 )
		<< "Average positive angle"    << sum1 * scale
		<< "Average negative angle"    << sum2 * scale
		<< "Average squared angle"	   << sum3 * scale
		<< "Average 10'th power angle" << sum4 * scale
		<< "Average 50'th power angle" << sum5 * scale;
}

/****************************************************************************
* Compute functionals based on arclength, directions, and positions.
****************************************************************************/
static void ArclengthFunctionals( const Multiline &lines, FeatureVector &F )
{
	const int N  = 12;
	Box2d box	 = lines.Box();
	double xmin	 = box.Xmin();
	double ymin	 = box.Ymin();
	double height = Max( box.Height(), 1.0 );
	double width  = Max( box.Width (), 1.0 );
	double length = lines.Length();
	double scale  = 1.0 / ( length + 1.0 );
	int   nvec	 = 0;
	Pnode *p	 = NULL;
	double Tmp[N];
	
	for( int i = 0; i < N; i++ ) Tmp[i] = 0.0;
	for( Pnode *q = lines.First(); q != NULL; q = q->Next() )
	{
		if( q->Jump() ) { p = q; continue; }
		nvec++;
		double  s   = q->Param() * scale;
		double  t   = 1.0 - s;
		double *T   = Tmp;
		Vec2   Q   = *q;
		double  fx  = ( Q.X() - xmin ) / width;
		double  fy  = ( Q.Y() - ymin ) / height;
		Vec2   V   = Q - *p;
		double  x   = V.X();
		double  y   = V.Y();
		
		// Compute functionals based on arclength from start.
		(*T++) += x * s;
		(*T++) += y * s;
		(*T++) += x * s * s;
		(*T++) += y * s * s;
		
		// Compute functionals based on arclength from end.
		(*T++) += x * t;
		(*T++) += y * t;
		(*T++) += x * t * t;
		(*T++) += y * t * t;
		
		// Compute functionals based on relative position.
		(*T++) += x * 6 * pow( fy - 0.5, 3 );
		(*T++) += y * 6 * pow( fx - 0.5, 3 );
		(*T++) += x * 4 * ( 1.0 - fy ) * fy;
		(*T++) += y * 4 * ( 1.0 - fx ) * fx;
		
		p = q;
	}
	
	// Insert the mins and maxes into the feature vector.
	F << "Arclength functionals";
	scale = 4.0 / ( lines.Length() + 1 );
	for( int j = 0; j < N; j++ ) F << Tmp[j] * scale;
}

/****************************************************************************
* Compute functionals based on multiple strokes.
****************************************************************************/
static void MultiStrokeFunctionals( const Multiline &lines, FeatureVector &F )
{
	double Size	  = lines.Diam();
	int   strokes = lines.Comps();
	double sum	  = 0.0;
	double sum2	  = 0.0;
	double max	  = 0.0;
	Vec2  Gap;
	if( strokes > 1 )
	{
		double p1, p2;
		Multiline Temp( lines );
		Multiline S1( CopyComponent( lines, 0 ) );
		for( int i = 1; i < strokes; i++ )
		{
			Multiline S2( CopyComponent( lines, i ) );
			Gap += ( S1.End() - S2.Start() );
			double dist = Distance( S1, S2, p1, p2 );
			sum  += dist;
			sum2 += dist * dist;
			if( dist > max ) max = dist;			
			S1 = S2;
		}			 
	}
	
	F << "Sum of gap vectors"
		<< Gap / Size;
	
	F << "Sum and sum of square distances between strokes"
		<< Vec2( sum, sum2 ) / Size;
	
	F << "Maximum distance between consecutive strokes"
		<< max;
}

/****************************************************************************
* Compute a number of basic featues and insert them into the feature vector
* (along with comment strings).  Also insert all the features computed by
* the other functions.
****************************************************************************/
void ExtractFeatures( const Multiline &lines, FeatureVector &F )
{
	double Size	  = lines.Diam();
	int   Points  = lines.Points();
	int   strokes = lines.Comps();
	Vec2  First   = lines.First()->Coord();
	Vec2  Last	  = lines.Last ()->Coord();
	Vec2  diag	  = lines.Vector();  // Vector from first coord to last.
	double width   = Max( lines.Width (), 1.0 );
	double height  = Max( lines.Height(), 1.0 );
	double rms	  = 0.0;
	double lmax	  = 0.0;
	
	F.Clear();
	
	for( int i = 0; i < strokes; i++ ) 
	{
		double len = lines.LengthOf(i);
		if( len > lmax ) lmax = len;
		rms += len * len;
	}
	rms = sqrt( rms / ( 0.25 * Size * Size ) );
	
	F << "Number of strokes (weighted)"
		<< 2 * lines.Comps();  
	
	F << "RMS segment lengths"
		<< rms;
	
	F << "Longest stroke normalized by size"
		<< lmax / Size;
	
	F << "Total length normalized by size"
		<< lines.Length() / Size;
	
	F << "Length of gaps normalized by size"
		<< GapLength( lines ) / Size;
	
	F << "Vector from start point to end point, normalized by size"
		<< diag / Size;
	
	F << "Starting angle"
		<< ( lines.Vertex(3) - lines.Vertex(0) ) / Size;
	
	// Fill in the remaining features using routines that each supply
	// a collection of related features.
	
	QuadrantHistogram	  ( lines, F );
	SectorHistogram 	  ( lines, F );
	AngleHistogram		  ( lines, F );
	ArclengthFunctionals  ( lines, F );
	MultiStrokeFunctionals( lines, F );
}

ostream &operator<<( ostream &out, const FeatureStats &s )
{
	char tmp[1000];
	sprintf(tmp, "%9.5f %9.5f %9.5f %9.5f\n", s.Average, s.StndDev, s.Min, s.Max);
	out << tmp;
	return out;
}

istream &operator>>( istream &in, FeatureStats &s )
{
	in >> s.Average
		>> s.StndDev
		>> s.Min
		>> s.Max;
	return in;
}

ostream &operator<<( ostream &out, const FeatureStatsVector &fsv )
{
	char buff[ 256 ];
	Symbol *symbol = fsv.GetSymbol();
	const char *sname = symbol->Name();  // Symbol name.
	const char *fname = symbol->Qual();  // File name.
	// Only keep the portion of the file name following "data/".
	if( getSubstringAfter( data_dir, fname, buff ) > 0 ) fname = buff;
	DirectoryTree::ConvertBackslashes(buff);
	out << sname << " " << fname << endl;
	for( register int j = 0; j < fsv.Size(); j++ ) 
	{
		if (j < 10) out << " ";
		out << j << ": " << fsv.Stats(j);
	}
	out << '.' << endl;
	return out;
}

istream &operator>>( istream &in, FeatureStatsVector &fsv )
{
	FeatureStats fs;
	char symb[64], file[128];
	// Read the symbol name and the name of the training file that
	// its samples were read from.
	in >> symb >> file;
	
	// Keep a pointer to the actual symbol, so that it can be looked up.
	fsv.SetDistance( 0.0 );
	fsv.SetSymbol( FontTable::FT.Insert( symb, file ) );
	
	for( register int j = 0; j < Max_Features; j++ ) 
	{
		int  num;
		char ch;
		in >> num; // Read (and ignore) the feature number.
		in >> ch;  // Read the colon.
		in >> fs;  // Read the actual statistics.
		in >> ch;  // Look for end sentinel.
		fsv << fs; // Add the feature stats to the feature stats vector.
		if( ch == '.' ) break; 
		in.putback( ch );
	}
	return in;
}
