/***************************************************************************
* Font.C																   *
*																		   *
* The "FontTable" class implements methods for reading font description    *
* files that have been generated from Encapsulated Postscript (EPS) files, *
* such as those generated by Adobe Illustrator.  The FontTable will 	   *
* contain information such as a bounding box per character (encoding	   *
* x-height and nominal width), skeleton, and positioning of superscripts   *
* and subscripts.  The resulting table can be queried to obtain polygonal  *
* outline fonts, width, superscript location, etc.						   *
*																		   *
*	Changes:															   *
*																		   *
*		01/04/2000	arvo	  Added GetMesh and GetTriangle.			   *
*		06/06/1998	arvo	  Replaced glut stroke characters.			   *
*		05/13/1998	arvo	  Added Greek letters.						   *
*		01/11/1998	arvo	  Started New Century Schoolbook font.		   *
*		07/14/1997	arvo	  Started symbol font.						   *
*		07/07/1997	arvo	  Made StrokeFont base class.				   *
*		07/03/1997	arvo	  Encode multiple strokes in one polyline.	   *
*		06/22/1997	novins	  Copied from smart/KeyPress.C				   *
*		09/04/1996	novins	  Initial coding.							   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
#include "stdafx.h"

#include <assert.h>
#include <windows.h>
#include "../util/Char.h"
#include "../util/Token.h"
#include "../util/DirectoryTree.h"
#include "../rec/Font.h"

FontTable FontTable::FT;

extern ofstream cerr, cout;

static const int num_sec = 10;
static Multiline Section[ num_sec ];

// Approximate a segment of a Bezier curve given by control points A, B, C, and D
// with a multiline.  The number of points in the multiline is computed using a 
// simple heuristic based on the length of the control polygon and the area of the
// triangles it forms.
//
Multiline Subdivide( const Vec2 &A, const Vec2 &B, const Vec2 &C, const Vec2 &D, int steps )
{
	Multiline approx;
	double d = dist( A, B ) + dist( B, C ) + dist( C, D );
	// double a1 = ( A - B ) ^ ( C - B );
	// double a2 = ( B - C ) ^ ( D - C );
	int N = 2 + (int)( steps * d + 0.5 );
	for( int i = 1; i <= N; i++ )
	{
		double t = double(i) / N;
		double s = 1.0 - t;
		Vec2 R1 = s * A  + t * B;
		Vec2 R2 = s * B  + t * C;
		Vec2 R3 = s * C  + t * D;
		Vec2 S1 = s * R1 + t * R2;
		Vec2 S2 = s * R2 + t * R3;
		Vec2 T1 = s * S1 + t * S2;
		approx << T1;
	}
	return approx;
}

static int Nested( const Multiline &A, const Multiline &B )
{
	return A.Inside( B.Start() ) || B.Inside( A.Start() );
}

// This function accepts a Multiline and interprets it as a collection
// of closed contours that represent one or more polygons with holes.
// What is returned is a multiline representing the same collection of 
// polygons, but without any holes.  That is, the holes are connected
// to the outer contour by a "cut line".  This is useful for drawing
// API's that do not support complex polygons (such as Tcl/Tk).
//
static void MakeSimple( Multiline &Pgons )
{
	int j, k = 0;
	Multiline Q;
	if( Pgons.Comps() <= 1 ) return;
	
	// Break up the components of the multiline into sections
	// consisting of nested polygons.
	Pgons.Pop( Section[0] );
	for(;;)
	{
		int spliced = 0;
		Pgons.Pop( Q );
		if( Q.Empty() ) break;
		for( j = 0; j <= k; j++ )
		{
			if( Nested( Q, Section[j] ) )
			{
				Section[j] = SpliceAtNearestPoint( Section[j], Q );
				spliced = 1;
				break;
			}
		}
		// If Q hasn't been spliced into another component, it
		// must be a disconnected component.
		if( !spliced ) 
		{
			assert( ++k < num_sec );
			Section[ k ] = Q;
		}
	}
	
	// Now copy all the sections back into the original multiline
	// as separate sections.
	Pgons.Clear();
	for( j = 0; j <= k; j++ ) 
	{
		Pgons += Section[j];
		Section[j].Clear();
	}
}

FontTable::FontTable()
{
	num_symbols  = 0;
	simple_pgons = 1;
	verbose 	 = NULL;
	progress	 = NULL;
}

FontTable::~FontTable()
{
	Clear();
}

void FontTable::Clear()
{
	num_symbols = 0;
	table.DeleteAll();
}

// Return a pointer to the complete symbol descriptor using
// only the name of the symbol.
//
Symbol *FontTable::Lookup( const char *name, const char *qual )
{
	return (Symbol *)table.Lookup( name, qual );
}

// Look up the given symbol name.  If it's not found, create
// a new symbol descriptor and insert it into the font table.
// In either case, set the "symbol" pointer so that subsequent
// operations can fill in fileds of the descriptor.
//
Symbol *FontTable::Insert( const char *symb, const char *fname )
{
	symbol = (Symbol *)table.Lookup( symb, fname );
	if( symbol == NULL )
	{
		symbol = new Symbol( symb, fname );
		table << symbol;
	}
	return symbol;
}

int FontTable::ReadDirectory( const char *path )
{
	DirectoryTree tree;
	tree.ScanDirectory(path);
	DT_File *file = tree.GetFirstFile();

	int count = 0;
	while (file != NULL) 
	{
		if (file->HasSuffix(".fdf"))
		{
			ReportProgress( "Loading font file", file->name);
			tok.Open(file->GetFullName());
			count += ReadFile();
		}
		file = tree.GetNextFile();
	}
	return count;
}

int FontTable::ReadFile()
{
	int count = 0;
	while( GetSymbol() ) count++;
	return count;
}

int FontTable::Error( const char *thing )
{
	cerr << "ERROR at line " << tok.Line()
		<< ", column "		<< tok.Column()
		<< " of file "		<< tok.FileName() 
		<< endl;
	if( thing != NULL ) 
		cerr << "      Expected " << thing << " and found \"" << tok.Spelling() << "\" instead." << endl;
	return 0;
}

int FontTable::GetSymbol()
{
	if( Null(++tok) ) return 0;
	if( tok == T_string ) Insert( tok.Spelling(), "<font>" );
	else return Error( "a string" );
	ReportProgress( "  symbol", metaString( tok.Spelling() ) );
	if( ++tok != Lcurley ) return Error( "a left brace" );
	for(;;)
	{
		if( GetWidth  () ) continue;
		if( GetStroke () ) continue;
		if( GetOutline() ) continue;
		if( GetPoints () ) continue;
		if( GetMesh   () ) continue;
		if( GetPos	  () ) continue;
		break;
	}
	if( ++tok != Rcurley ) return Error( "a right brace" );
	
	// Now fill in any missing information with reasonable 
	// approximations based on the raw character geometry.
	symbol->ComputeMissingFields();
	num_symbols++;
	return 1;
}

int FontTable::GetWidth()
{
	if( ++tok != "width"   ) { --tok; return 0; }
	if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
	symbol->SetWidth( tok.Fvalue() );
	return 1;
}

int FontTable::GetStroke()
{
	if( ++tok != "stroke" ) { --tok; return 0; }
	if( !GetLine()		  ) {		 return 0; }
	symbol->SetStroke( line );
	return 1;
}

int FontTable::GetOutline()
{
	if( ++tok != "outline" ) { --tok; return 0; }
	if( !GetLine()		   ) {		  return 0; }
	if( simple_pgons ) MakeSimple( line );
	symbol->SetOutline( line );
	return 1;
}

int FontTable::GetPoints()
{
	if( ++tok != "points" ) { --tok; return 0; }
	if( ++tok != Lcurley  ) { --tok; return Error( "a left brace" ); }
	points.Clear();
	while( GetCoord() );
	if( ++tok != Rcurley  ) { --tok; return Error( "a right brace" ); }
	return 1;
}

int FontTable::GetCoord()
{
	double x, y, w = 0.0;
	if( ++tok != T_numeric ) { --tok; return 0; }
	x = tok.Fvalue();
	if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
	y = tok.Fvalue();
	if( ++tok == T_numeric ) 
	{
		w = tok.Fvalue();
		++tok;
	}
	if( tok != Comma ) --tok; // Skip over commas.	Put everything else back.
	// Flip the Y-coordinate so that Y is down (to match Tcl).
	points	<< TimedVec2( x, -y, (long)(100 * w) );
	return 1;
}

int FontTable::GetMesh()
{
	if( ++tok != "mesh"  ) { --tok; return 0; }
	if( ++tok != Lcurley ) { --tok; return Error( "a left brace" ); }
	mesh.Clear();
	while( GetTriangle() );
	if( ++tok != Rcurley  ) { --tok; return Error( "a right brace" ); }
	symbol->SetMesh( mesh );
	return 1;
}

int FontTable::GetTriangle()
{
	double x, y, w;
	for( int i = 0; i < 3; i++ )
	{
		if( ++tok != T_numeric ) { --tok; return i == 0 ? 0 : Error( "a number" ); }
		x = tok.Fvalue();
		if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
		y = tok.Fvalue();
		if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
		w = tok.Fvalue();
		// NOTE: Using the time field of a TimedVec2 is a hack for encoding intensity.
		// This should be changed to an appropriate data structure at some point.
		mesh << TimedVec2( x, -y, (long)(1000 * w) );  
	}
	// Encode each triangle as a separate sequence of three vertices within
	// a multiline.
	mesh << Multiline::Jump;
	return 1;
}

// Read in the superscrip and subscript positions; they will be left as
// "NotSupplied" if they are not found.  In the latter case, they will
// be filled in with a reasonable guess at the very end, when all
// information is sure to be available (such as the bounding box
// of the outline representation).
//
int FontTable::GetPos()
{
	int sub;
	++tok;
	if( tok == "sub_pos" ) sub = 1; else
		if( tok == "exp_pos" ) sub = 0; else { --tok; return 0; }
		if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
		double x = tok.Fvalue();
		if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
		double y = tok.Fvalue();
		// Flip the Y-coordinate so that Y is down (to match Tcl).
		if( sub ) 
			symbol->SetSubPos( Vec2( x, -y ) );
		else symbol->SetExpPos( Vec2( x, -y ) );
		return 1;
}

int FontTable::GetLine()
{
	if( ++tok != Lcurley  ) { --tok; return Error( "a left brace"  ); }
	line.Clear();
	for(;;)
	{
		if( GetPoint() ) continue;
		if( GetCurve() ) continue;
		break;
	}
	if( ++tok != Rcurley  ) { --tok; return Error( "a right brace" ); }
	return 1;
}

int FontTable::GetPoint()
{
	int jump;
	++tok;
	if( tok == 'm' ) jump = 1; else
		if( tok == 'l' ) jump = 0; else { --tok; return 0; }
		if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
		double x = tok.Fvalue();
		if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
		double y = tok.Fvalue();
		if( jump ) line << Multiline::Jump;
		// Flip the Y-coordinate so that Y is down (to match Tcl).
		line << Vec2( x, -y );
		return 1;
}

int FontTable::GetCurve()
{
	double c[6];
	if( ++tok != 'c' ) { --tok; return 0; }
	for( int i = 0; i < 6; i++ )
	{
		if( ++tok != T_numeric ) { --tok; return Error( "a number" ); }
		c[i] = tok.Fvalue();
	}
	line << Subdivide( 
		line.End(),
		Vec2( c[0], -c[1] ), 
		Vec2( c[2], -c[3] ), 
		Vec2( c[4], -c[5] ) 
		);
	return 1;
}

// Look up the string in the global font table and return the corresponding
// symbol.	If the string is actually a "meta string", with literal quotes
// included, they are stripped off before doing the lookup.
//
Symbol *GetSymbolPtr( const char *str )
{
	Symbol *sym;
	if( str[0] == DoubleQuote && str[1] != NullChar )
		sym = FontTable::FT.Lookup( stripQuotes( str ) );
	else sym = FontTable::FT.Lookup(			  str	);
	return sym;
}

void FontTable::ReportProgress( const char *msg, const char *data )
{
	char buff[1024];
	sprintf( buff, "%s %s", msg, data );
	if( progress != NULL ) progress( buff );
	if( verbose  != NULL ) *verbose << buff << endl;
}
