/***************************************************************************
* Font.h                                                                   *
*                                                                          *
* The "FontTable" class implements methods for reading and querying font   *
* descriptions that have been generated by "ps2fdf" from Encapsulated      *
* Postscript (EPS) files, which in turn, were generated by Adobe           *
* Illustrator.                                                             *
*                                                                          *
* The "font description files" (.fdf) contain information such as bounding *
* boxes (encoding x-height and nominal width), skeleton, and positioning   *
* of superscripts and subscripts for each character.  The resulting table  *
* can be queried to obtain polygonal outline fonts, width, superscript     *
* location, etc.                                                           *
*                                                                          *
*   Changes:                                                               *
*                                                                          *
*       01/04/2000  arvo      Added GetMesh and GetTriangle.               *
*       05/13/1998  arvo      Added Greek letters.                         *
*       07/07/1997  arvo      Made StrokeFont base class.                  *
*       06/22/1997  novins    Initial coding.                              *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#ifndef __FONT_INCLUDED__
#define __FONT_INCLUDED__

#include "../util/Token.h"
#include "../util/Name.h"
#include "../rec/Symbol.h"

typedef void (*progress_f)( const char *msg );

class FontTable {

    public:
        static FontTable FT; // A globally accessible table.

    public:
        FontTable();
       ~FontTable();
        void Clear();
        Symbol *Lookup( const char *name, const char *qual = NULL );
        int ReadDirectory( const char *path );
        Symbol *Insert( const char *symb, const char *qual = NULL );
        void Verbose( ostream &out ) { verbose = &out; }
        void SetProgressCallback( progress_f func ) { progress = func; }
        int NumSymbols() const { return num_symbols; }
        void SimplifyPolygons( int yes_no ) { simple_pgons = yes_no; }

    private:
        int ReadFile   ();
        int GetSymbol  ();
        int GetWidth   ();
        int GetStroke  ();
        int GetOutline ();
        int GetPoints  ();
        int GetMesh    ();
        int GetTriangle();
        int GetPos     ();
        int GetPoint   ();
        int GetCurve   ();
        int GetLine    ();
        int GetCoord   ();
        int Error( const char *thing = NULL );

    private:
        void       ReportProgress( const char *, const char * );
        int        num_symbols;
        int        simple_pgons;
        ostream   *verbose;
        Symbol    *symbol;
        Token      tok;
        NameTable  table;
        Multiline  line;
        progress_f progress;
        Multiline  points;   // For point lists.
        Multiline  mesh;     // For triangular meshes.
        Vec2       start;    // Start of a line or curve.
    };


// Look up the string in the global font table and return the corresponding
// symbol.  If the string is actually a "meta string", with literal quotes
// included, they are stripped off before doing the lookup.
extern Symbol *GetSymbolPtr( 
    const char *str 
    );

// Approximate a segment of a Bezier curve given by control points A, B, C, and D
// with a multiline.  The number of points in the multiline is computed using a 
// simple heuristic based on the length of the control polygon.
extern Multiline Subdivide( 
    const Vec2 &, 
    const Vec2 &, 
    const Vec2 &, 
    const Vec2 &,
    int steps = 10  // Controls level of subdivision.
    );

#endif








