/***************************************************************************
* LineMorph.C															   *
*																		   *
* This module defines all the functions that are called from Tcl/Tk via    *
* the SendMessage and Dispatch mechanisms.								   *
*																		   *
*	Changes:															   *
*																		   *
*		10/09/99	arvo	Integrated into Train program.				   *
*		06/22/97	arvo	Rewritten for morph project.				   *
*		07/06/97	arvo	Initial coding. 							   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
#include <assert.h>
#include "../rec/LineMorph.h"

// Destructively update the Multiline P by introducing a break
// at the largest bend whose angle is no more than "max_angle".
// The "angle" is between 0 and 1, where 0 is a line that is exactly
// folded over on itself, and 1 is an unbent line.
//
int BreakSharpestBend( Multiline &P, double max_angle )
{
	if( P.Points() < 7 ) return 0;
	double  min_angle = max_angle;
	Pnode *min_node = NULL;
	int skip = 0;
	
	Pnode *n3 = P.First();
	Pnode *n2 = n3->Next(); if( n2->Jump() ) skip = 1;
	Pnode *n1 = n2->Next(); if( n1->Jump() ) skip = 2;
	Pnode *c  = n1->Next(); if( c ->Jump() ) skip = 3;
	Pnode *p1 = c ->Next(); if( p1->Jump() ) skip = 4;
	Pnode *p2 = p1->Next(); if( p2->Jump() ) skip = 5;
	Pnode *p3 = p2->Next();
	while( p3 != NULL )
	{
		if( p3->Jump() ) skip = 6;
		if( skip-- <= 0 )
		{
			Vec2 A = *n3 - *c;
			Vec2 B = *p3 - *c;
			if( LenSqr( A ) > 0 && LenSqr( B ) > 0 )
			{
				double angle = ArcCos( Unit(A) * Unit(B) ) / Pi;
				if( angle <= min_angle )
				{
					min_angle = angle;
					min_node  = c;
				}
			}
		}
		// Shift all the points down by one.
		n3 = n2;
		n2 = n1;
		n1 = c;
		c  = p1;
		p1 = p2;
		p2 = p3;
		p3 = p3->Next();
	}
	if( min_node == NULL ) return 0;
	
	// Now construct a new multiline that has a break at the
	// designated node.  Double the corresponding vertex so that
	// no gap is introduced.
	
	Multiline Temp;
	for( Pnode *p = P.First(); p != NULL; p = p->Next() )
	{
		if( p->Jump() ) Temp << Multiline::Jump;
		Temp << *p;
		if( p == min_node ) 
		{
			// Introduce a jump, and double the vertex.
			Temp << Multiline::Jump;
			Temp << *p;
		}
	}
	P = Temp;
	return 1;
}


// Refine the multiline P by inserting new points along it that correspond to 
// the positions of the vertices in Q.	All parametric distances are normalized 
// so that the parameter of the last point is always 1.  This function creates 
// a new multiline and returns a pointer to it.  This only works for single
// connected multilines!
//
static Multiline SimpleRefine( const Multiline &P, const Multiline &Q )
{
	Multiline R;
	if( Q.Points() < 2 ) return P;
	if( P.Points() < 2 ) 
	{
		for( int i = 0; i < Q.Points() - P.Points(); i++ )
			R << Q.Start();
		return R;
	}
	
	// Define a tolerance for points that are to be assumed equal.
	
	const double epsilon = 1.0E-5; 
	
	// p_sacle and q_scale will be used to compute the
	// normalized parametric distances.
	
	double p_scale = 1.0 / P.Length();
	double q_scale = 1.0 / Q.Length();
	
	// Denote the first segment of P as (p0,p1), and the
	// first segment of Q as (q0,q1).
	
	Pnode *p0 = P.First(), *p1 = p0->Next();
	Pnode *q0 = Q.First(), *q1 = q0->Next();
	
	// p_next and q_next denote the parametric distance to the
	// next vertex of P and of Q respectively.
	
	double p_next = p_scale * p1->Param();
	double q_next = q_scale * q1->Param();
	
	R << P.First()->Coord();
	while( p1 != NULL )
	{
		while( q1 != NULL && q_next < p_next - epsilon )
		{
			R << Interpolate( p0, p1, q_next * P.Length() );
			q0 = q1;
			q1 = q1->Next();
			q_next = q1 == NULL ? 1.0 : q_scale * q1->Param(); 
		}
		if( Abs( q_next - p_next ) > epsilon ) R << p1->Coord();
		p0 = p1;
		p1 = p1->Next();
		p_next = p1 == NULL ? 1.0 : p_scale * p1->Param(); 
	}
	R << P.Last()->Coord();
	return R;
}

// Sample both Multilines at N evenly-spaced intervals and compute the sum of
// the distances between the corresponding paris, parametrized in both directions.
// Return the two sums as "s" and "t".
//
static inline void Energy( const Multiline &A, const Multiline &B, double &s, double &t )
{
	const int N = 7;
	s = 0;
	t = 0;
	Vec2 a[N], b[N];
	for( int i = 0; i < N; i++ )
	{
		double t = double(i) / (N-1);
		a[i] = A(t);
		b[i] = B(t);
	}
	for( int j = 0; j < N; j++ )
	{
		int k = N - j - 1;
		s += dist( a[j], b[j] );
		t += dist( a[j], b[k] );
	}
	s /= N;
	t /= N;
}

double TotalEnergy( const Multiline &P, const Multiline &Q )
{
	// Compute the total energy needed to transform the components of
	// P into Q, or vise versa.  If one multiline has more components than
	// the other, then some components will be skipped.
	Multiline A(P), B(Q);
	Multiline S1, S2;
	double s, t, energy = 0.0;
	while( !A.Empty() && !B.Empty() )
	{
		A.Pop( S1 );
		B.Pop( S2 );
		Energy( S1, S2, s, t );
		energy += Min( s, t );
	}
	return energy / Max( P.Comps(), Q.Comps(), 1 );
}

static void Flip_to_Minimize( Multiline &A, const Multiline &B )
{
	double s, t;
	Energy( A, B, s, t );
	if( s > t ) A.Reverse();
}

static void operator|=( Multiline &P, const Multiline &Q )
{
	Multiline A( P ), B( Q );
	Multiline S1, S2, Result;
	while( !A.Empty() && !B.Empty() )
	{
		A.Pop( S1 );
		B.Pop( S2 );
		Result += SimpleRefine( S1, S2 );
	}
	P = Result;
}


// Permute and flip the components of P so that they are
// in the best 1-1 correspondence with Q.  That is, order
// them so that the sum of the distortion energies is minimized.
// This is currently an O(N!) algorithm, where N is the number 
// of components.

static double MatchComponents( Multiline &P, const Multiline &Q )
{
	assert( P.Comps() >= Q.Comps() );
	
	Perm best( 0, P.Comps() - 1 ), perm( 0, P.Comps() - 1 );
	Multiline A;
	double energy = TotalEnergy( P, Q );
	for(;;)
	{
		A = P;
		if( ++perm == 0 ) break;
		double e = TotalEnergy( A *= perm, Q );
		if( e < energy )
		{
			best   = perm;
			energy = e;
		}
		
	}
	P *= best;
	
	Multiline T( Q );
	Multiline S1, S2, Result;
	while( !P.Empty() )
	{
		P.Pop( S1 );
		if( !T.Empty() )  
		{
			T.Pop( S2 );
			Flip_to_Minimize( S1, S2 );
		}
		Result += S1;
	}
	P = Result;
	return energy;
}


// Both multilines are modified by splitting, refining, reordering
// and reparameterizing (flipping) their components so that
//	  1) they have the same number of comonents
//	  2) corresponding components have the same number of points
//	  3) total energy of the matching is minimized.
//
double LineMorphSetup( Multiline &P, Multiline &Q )
{
	const double max_angle = 0.75;
	double energy;
	
	// Force the multilines to have the same number of components.
	// Introduce more components into the one with the fewest by
	// splitting at the sharpest bends.
	
	while( P.Comps() < Q.Comps() )
	{
		if( !BreakSharpestBend( P, max_angle ) ) break;
	}
	
	while( Q.Comps() < P.Comps() )
	{
		if( !BreakSharpestBend( Q, max_angle ) ) break;
	}
	
	// If the above failed introduce enough breaks so that the two
	// multilines have the same number of components, we reorder and/or 
	// reverse some of the segments of P or Q to produce as good a
	// matching of the components that exist as possible.  Then we
	// add some degenerate line segments to "pad" the one with the 
	// fewest components.
	if( P.Comps() >= Q.Comps() ) 
		energy = MatchComponents( P, Q );
	else energy = MatchComponents( Q, P );
	
	// Add degenerate components to P, if necessary.
	while( P.Comps() < Q.Comps() )
	{
		int  k	= P.Comps();
		Vec2 A1 = Q.StartOf( k );
		Vec2 A2 = P.Closest( A1 );
		Vec2 B1 = Q.EndOf( k );
		Vec2 B2 = P.Closest( B1 );
		double da = dist(A1,A2);
		double db = dist(B1,B2);
		P << Multiline::Jump;
		if( RelErr( da, db ) < 0.01 ) P << CopyComponent(Q,k)(0.5);
		else if( da < db ) P << A2;
		else P << B2;
		P << P.End() + Vec2(1,0);
	}
	
	// Add degenerate components to Q, if necessary.
	while( Q.Comps() < P.Comps() )
	{
		int  k	= Q.Comps();
		Vec2 A1 = P.StartOf( k );
		Vec2 A2 = Q.Closest( A1 );
		Vec2 B1 = P.EndOf( k );
		Vec2 B2 = Q.Closest( B1 );
		double da = dist(A1,A2);
		double db = dist(B1,B2);
		Q << Multiline::Jump;
		if( RelErr( da, db ) < 0.01 ) Q << CopyComponent(P,k)(0.5);
		else if( da < db ) Q << A2;
		else Q << B2;
		Q << Q.End() + Vec2(1,0);
	}
	
	// Finally, refine each pair of segments so that their vertices are
	// in 1-1 correspondence and their normalized parameters match.
	// This is done at the end, since it is needed only for the interpolation
	// step -- it has nothing at all to do with energy minimization.
	SimpleMorphSetup( P, Q );
	
	return energy;
}

Multiline Interpolate( double t, const Multiline &P, const Multiline &Q )
{
	Multiline R;
	Pnode *p = P.First();
	Pnode *q = Q.First();
	while( p != NULL && q != NULL )
	{
		if( p->Jump() || q->Jump() ) R << Multiline::Jump;
		R << Interpolate( p->Coord(), q->Coord(), t );
		p = p->Next();
		q = q->Next();
	}
	return R;
}


void SimpleMorphSetup( Multiline &P, Multiline &Q ) 
{
	assert( P.Comps() == Q.Comps() );
	Multiline T( P );
	P |= Q; // Refine P against Q.
	Q |= T; // Refine Q against the old P.
}

#include "stdafx.h"
