/***************************************************************************
* Box2.C																   *
*																		   *
* Basic 2D bounding boxes.												   *
*																		   *
*	HISTORY 															   *
*		05/12/98	arvo	Broken out into a separate module.			   *
*		12/04/97	arvo	Reorganized for use as a library.			   *
*		07/09/96	arvo	Initial coding. 							   *
*																		   *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo										   *
*																		   *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html	   *
*																		   *
* This program is distributed in the hope that it will be useful, but	   *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more	   *
* details.																   *
*																		   *
***************************************************************************/
#include "stdafx.h"

#include <stdio.h>
#include "../util/Char.h"
#include "../util/Box2.h"

const Box2d Box2d::Zero(	0.0    );
const Box2d Box2d::Huge(  Infinity );
const Box2d Box2d::Null( -Infinity );

Box2d::Box2d()
{
	// By default, all boxes are maximally "degenerate".  That is,
	// the minimum values are "infinity" (some huge number) while the
	// the maximum values are negative "infinity".	Any expansion
	// operation on such a box (such as inserting a point or another 
	// box) will update all the entries.
	//
	*this = Box2d::Null;
}

Box2d::Box2d( double size )
{
	// Create a box of the given size.	If size < 0, this will create
	// a "degenerate" box consisting of two intervals with min greater
	// than max.  This constructor can ONLY be used to initialize
	// the static data members of the Box2d class.
	//
	xmin = -size;
	xmax =	size;
	ymin = -size;
	ymax =	size;
}

Box2d &Box2d::Set( const Vec2 &A, const Vec2 &B )
{
	// Create a bounding box from two opposing corners.
	// This is equivalent to *this << A << B.
	//
	if( A.X() <= B.X() )
	{ 
		xmin = A.X(); 
		xmax = B.X(); 
	}
	else 
	{ 
		xmin = B.X(); 
		xmax = A.X(); 
	}
	if( A.Y() <= B.Y() )
	{ 
		ymin = A.Y(); 
		ymax = B.Y(); 
	}
	else 
	{ 
		ymin = B.Y(); 
		ymax = A.Y(); 
	}
	return *this;
}

Box2d &Box2d::operator+=( const Vec2 &p )  
{
	// Return a new Translated box.
	//
	xmin += p.X();
	xmax += p.X();
	ymin += p.Y();
	ymax += p.Y();
	return *this;
}

Box2d &Box2d::operator-=( const Vec2 &p )  
{
	// Translate this box.
	//
	xmin -= p.X();
	xmax -= p.X();
	ymin -= p.Y();
	ymax -= p.Y();
	return *this;
}

Box2d Box2d::operator*( double scale ) const
{
	// Return a new scaled box.
	//
	return Box2d( scale * Min(), scale * Max() );
}

Box2d &Box2d::operator*=( double scale )  
{
	// Scale this box.
	//
	xmin *= scale;
	xmax *= scale;
	ymin *= scale;
	ymax *= scale;
	return *this;
}

Box2d &Box2d::operator^=( double scale )  
{
	// Scale this box about its center.
	//
	Vec2 C = Center();
	xmin = scale * ( xmin - C.X() ) + C.X();
	xmax = scale * ( xmax - C.X() ) + C.X();
	ymin = scale * ( ymin - C.Y() ) + C.Y();
	ymax = scale * ( ymax - C.Y() ) + C.Y();
	return *this;
}

Box2d &Box2d::operator<<( const Vec2 &p )  
{
	// Expand this box to include another point.
	//
	if( p.X() > xmax ) xmax = p.X();
	if( p.X() < xmin ) xmin = p.X();
	if( p.Y() > ymax ) ymax = p.Y();
	if( p.Y() < ymin ) ymin = p.Y();
	return *this;
}

Box2d &Box2d::operator<<( const Box2d &b )	
{
	// Expand this box to include another box.
	//
	if( b.Xmax() > xmax ) xmax = b.Xmax();
	if( b.Xmin() < xmin ) xmin = b.Xmin();
	if( b.Ymax() > ymax ) ymax = b.Ymax();
	if( b.Ymin() < ymin ) ymin = b.Ymin();
	return *this;
}

Box2d Box2d::operator+( const Vec2 &T ) const 
{
	// Return a new translated box.
	//
	return Box2d( Vec2(xmin,ymin) + T, Vec2(xmax,ymax) + T );
}

Box2d Box2d::operator-( const Vec2 &T ) const 
{
	// Return a new translated box.
	//
	return Box2d( Vec2(xmin,ymin) - T, Vec2(xmax,ymax) - T );
}

Box2d &Box2d::operator*=( const Box2d &B )
{
	// Shrink this box to be inside a second.  That is, set this
	// box to be the intersection of itself with a second box.
	//
	xmin = ::Max( xmin, B.Xmin() );
	xmax = ::Min( xmax, B.Xmax() );
	ymin = ::Max( ymin, B.Ymin() );
	ymax = ::Min( ymax, B.Ymax() );
	if( xmin > xmax || ymin > ymax ) *this = Box2d::Null;
	return *this;
}

Box2d Box2d::operator*( const Box2d &B ) const
{
	// Create a new box that is the intersection of this box with
	// a second box.
	//
	return Box2d( *this ) *= B;
}

double Box2d::Diam() const
{
	// Return the "diameter" of the box, which is zero if the box
	// is degenerate, and the distance between opposing corners otherwise.
	//
	if( Xmin() > Xmax() ) return 0.0;
	if( Ymin() > Ymax() ) return 0.0;
	return ::dist( SW(), NE() );
}

double Box2d::Area() const
{
	// Return the area of the box, or zero if the box is degenerate.
	//
	if( Xmin() >= Xmax() ) return 0.0;
	if( Ymin() >= Ymax() ) return 0.0;
	return ( Xmax() - Xmin() ) * ( Ymax() - Ymin() );
}

double Box2d::Perim() const
{
	// Return the perimeter of the box, or zero if the box is degenerate.
	//
	if( Xmin() > Xmax() ) return 0.0;
	if( Ymin() > Ymax() ) return 0.0;
	return 2.0 * ( Height() + Width() );
}

Box2d Box2d::Expand( double delta ) const
{
	// Return a new box that is expanded by "delta" in all directions.
	//
	Vec2 D( delta, delta );
	return Box2d( Min() - D, Max() + D );
}

double Box2d::dist( const Vec2 &p ) const
{
	// Compute the 2-Norm distance between this box and a point.
	// The distance is zero if the point is in or on the box.
	//
	double d = 0.0;
	
	if( p.X() > Xmax() ) d += Sqr( p.X() - Xmax() ); else
		if( p.X() < Xmin() ) d += Sqr( p.X() - Xmin() );
		
		if( p.Y() > Ymax() ) d += Sqr( p.Y() - Ymax() ); else
			if( p.Y() < Ymin() ) d += Sqr( p.Y() - Ymin() );
			
			return sqrt( d );
}

double Box2d::dist( const Box2d &B ) const
{
	// Compute the 2-Norm distance between this box and another box.
	// The distance is zero if the boxes overlap.
	//
	double d = 0.0;
	
	if( B.Xmin() > Xmax() ) d += Sqr( B.Xmin() - Xmax() ); else
		if( B.Xmax() < Xmin() ) d += Sqr( B.Xmax() - Xmin() );
		
		if( B.Ymin() > Ymax() ) d += Sqr( B.Ymin() - Ymax() ); else
			if( B.Ymax() < Ymin() ) d += Sqr( B.Ymax() - Ymin() );
			
			return sqrt( d );
}

Vec2 Box2d::Closest( const Box2d &B ) const
{
	// Return the shortest vector from a point on this box to the other box.
	// The vector is zero if the boxes overlap.
	//
	Vec2 v;
	
	if( B.Xmin() > Xmax() ) v.X() = B.Xmin() - Xmax(); else
		if( B.Xmax() < Xmin() ) v.X() = B.Xmax() - Xmin();
		
		if( B.Ymin() > Ymax() ) v.Y() = B.Ymin() - Ymax(); else
			if( B.Ymax() < Ymin() ) v.Y() = B.Ymax() - Ymin();
			
			return v;
}

Vec2 Box2d::Closest( const Vec2 &p ) const
{
	// Return the point on the boundary of this box that is   
	// closest to the point p.		
	//
	double x, y;
	int   in_x = 0, in_y = 0;
	
	if( p.X() > Xmax() ) x = Xmax(); else
		if( p.X() < Xmin() ) x = Xmin(); else { in_x = 1; x = p.X(); }
		
		if( p.Y() > Ymax() ) y = Ymax(); else
			if( p.Y() < Ymin() ) y = Ymin(); else { in_y = 1; y = p.Y(); }
			
			// If p ended up inside the box, find the coordinate that is
			// closest to the edge and set that coordinate to the boundary
			// value (either min or max, whichever is closer).
			
			if( in_x && in_y )
			{
				double x0 = Abs( x - Xmin() );
				double x1 = Abs( x - Xmax() );
				double y0 = Abs( y - Ymin() );
				double y1 = Abs( y - Ymax() );
				if( ::Min( x0, x1 ) < ::Min( y0, y1 ) )
					x = ( x0 < x1 ) ? Xmin() : Xmax();
				else y = ( y0 < y1 ) ? Ymin() : Ymax();
			}
			
			return Vec2( x, y );
}

int Box2d::Inside( const Vec2 &p ) const
{
	// Return 1 if the point "p" is in or on the box, and 0 otherwise.
	//
	if( p.X() < Xmin() || Xmax() < p.X() ) return 0;
	if( p.Y() < Ymin() || Ymax() < p.Y() ) return 0;
	return 1;
}

double dist( const Box2d &A, const Box2d &B )
{
	// Compute the 2-Norm distance between two boxes, which is zero
	// if the boxes intersect.
	//
	double dist = 0.0;
	
	if( B.Xmin() > A.Xmax() ) dist += Sqr( B.Xmin() - A.Xmax() ); else
		if( A.Xmin() > B.Xmax() ) dist += Sqr( A.Xmin() - B.Xmax() );
		
		if( B.Ymin() > A.Ymax() ) dist += Sqr( B.Ymin() - A.Ymax() ); else
			if( A.Ymin() > B.Ymax() ) dist += Sqr( A.Ymin() - B.Ymax() );
			
			return sqrt( dist );
}

int hit( const Box2d &A, const Box2d &B )
{
	// Return 1 if boxes A and B intersect (even at just an edge or corner), 
	// and return 0 otherwise.
	//
	if( B.Xmin() > A.Xmax() ) return 0;
	if( B.Ymin() > A.Ymax() ) return 0;
	if( A.Xmin() > B.Xmax() ) return 0;
	if( A.Ymin() > B.Ymax() ) return 0;
	return 1;
}

ostream &operator<<( ostream &out, const Box2d &B )
{
	// Write the box intervals and area to standard out.
	//
	out << "(" << B.Xmin() << "," << B.Xmax() << ")x"
		<< "(" << B.Ymin() << "," << B.Ymax() << ")"
		<< endl;
	return out;
}

const char *Box2d::ToString( char *buff ) const
{
	// Return a string that encodes the box exactly -- writing out the
	// binary representation of the floating point values in hexadecimal.
	//
	static char local_buff[64];
	if( buff == NULL ) buff = local_buff;
	sprintf( buff, "%lX %lX %lX %lX", 
		float_as_long( Xmin() ), 
		float_as_long( Ymin() ), 
		float_as_long( Xmax() ), 
		float_as_long( Ymax() )
		);
	return buff;
}

Box2d &Box2d::FromString( const char *str )
{
	// Read back the hexadecimal encoding of the box.
	//
	long i_xmin, i_xmax, i_ymin, i_ymax;
	sscanf( str, "%lX %lX %lX %lX", &i_xmin, &i_ymin, &i_xmax, &i_ymax );
	xmin = long_as_float( i_xmin );
	ymin = long_as_float( i_ymin );
	xmax = long_as_float( i_xmax );
	ymax = long_as_float( i_ymax );
	return *this;
}
