/***************************************************************************
* Multiline.h                                                              *
*                                                                          *
* The Multiline class encapsulates sequences of 2D points with time-       *
* stamps; the sequences encode one or more "polylines", which are chains   *
* of line segments.  The class automatically maintains information such    *
* as the bounding box, length, number of disconnected segments, etc.       *
*                                                                          *
*   HISTORY                                                                *
*     10/09/99    arvo    Added Resample function.                         *
*     08/09/98    arvo    Added Distance, Split, and Splice.               *
*     05/14/98    arvo    Converted Polyline to Multiline.                 *
*     05/12/98    arvo    Split off into a separate module.                *
*     12/04/97    arvo    Reorganized for use as a library.                *
*     07/07/97    arvo    Added TimedVec2 class.                           *
*     07/03/97    arvo    Multilines can now encode jumps.                 *
*     06/17/97    arvo    Added Multiline code.                            *
*     07/09/96    arvo    Initial coding.                                  *
*                                                                          *
*--------------------------------------------------------------------------*
* Copyright (C) 1999, James Arvo                                           *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
***************************************************************************/
#ifndef __MULTILINE_INCLUDED__
#define __MULTILINE_INCLUDED__

//#include <iostream.h>
#include "../util/Vec2.h"   // 2-D vectors.
#include "../util/Box2.h"   // 2-D boxes.
#include "../util/Trans2.h" // 2-D affine transformations.
#include "../util/Perm.h"   // Permutations.

class Pnode    ; // Used for creating vertex lists in a Multiline.
class Multiline; // Sequence of connected 2-D vertices (with jumps).

// The Pnode class is used for creating the lists of 2-D vertices 
// that define Multilines.  Only the Multiline class can create and
// initialize these nodes.
//
class Pnode : public TimedVec2 {

    public:
        Pnode *Next  () const { return next ; }
        double  Param () const { return param; }
        int    Jump  () const { return param == 0.0; }
       ~Pnode() {}

    private:
        Pnode() {}    // Only Multiline class has access to constructor.
        Pnode *next;  // Pointer to next node in the list.
        double  param; // Distance along Multiline. Zero means new segment.

    friend class Multiline;
    };


// The Multiline class encapsulates a sequence of 2D points representing one
// or more sequences of connected lines, separated by jumps.  A bounding box
// containing all vertices is automatically maintained.
//
class Multiline {

    public: // Constructors and destructor.
        Multiline();                     // Makes an empty Multiline.
        Multiline( const Multiline &P ); // Copy constructor.
        Multiline( const Multiline *P ); // Copy constructor.
        Multiline( const Box2d     &B ); // Makes a Multiline for drawing a box.
       ~Multiline();                     // Garbage collects the vertices.

    public: // Const operators.
        Vec2 operator()( double u ) const; // Parametrization over [0,1].

    public: // Non-const operators.
        Multiline &operator<<( const Vec2      & ); // Append new vertex.
        Multiline &operator<<( const TimedVec2 & ); // Append new vertex with time stamp.
        Multiline &operator+=( const Vec2      & ); // Translate.
        Multiline &operator-=( const Vec2      & ); // Translate.
        Multiline &operator*=(       double       ); // Scale.
        Multiline &operator*=( const Mat2x2    & ); // Transform all vertices.
        Multiline &operator*=( const Trans2    & ); // Transform all vertices.
        Multiline &operator/=( const Trans2    & ); // Transform all vertices.
        Multiline &operator^=(       double       ); // Centered scale.
        Multiline &operator =( const Multiline & ); // Copy another Multiline.
        Multiline &operator =( const Box2d     & ); // Make a Multiline from a box.
        Multiline &operator<<( const Multiline & ); // Append to last component.
        Multiline &operator+=( const Multiline & ); // Append a new component.
        Multiline &operator*=( const Perm      & ); // Permute the components.

    private: // Static data for maintaining global statistics & free lists.
        static Pnode *free_node; // Free list for link nodes.
        static long   num_nodes; // Number of link nodes allocated.
        static long   num_lines; // Number of Multilines allocated.

    public: // Const member functions that return basic info.
        Box2d  Box     () const;                    // Box around all vertices.
        Box2d  TrueBox ();                          // Box around all vertices.
        Pnode *First   () const { return  first ; } // First vertex.
        Pnode *Last    () const { return  last  ; } // Last vertex.
        Pnode *Penult  () const { return  penult; } // Next-to-last vertex.
        double  Length  () const { return  length; } // Total length of line.
        int    Points  () const { return  points; } // Number of vertices.
        int    Comps   () const { return  comps ; } // Number of components.
        int    Weight  () const { return  weight; } // Line weight.
        int    Empty   () const { return !points; } // True if no points.
        Vec2   Start   () const;                    // First vertex.
        Vec2   End     () const;                    // Last vertex.
        long   Started () const;                    // Time of first vertex.
        long   Ended   () const;                    // Time of last vertex.
        double  Height  () const { return box.Height(); }
        double  Width   () const { return box.Width (); }
        double  Diam    () const { return box.Diam  (); }
        double  Area    () const { return box.Area  (); }
        int    Compound() const { return comps > 1;    }
        Vec2   Center  () const { return box.Center(); }
        Vec2   Vector  () const { return End() - Start(); }
        Vec2   Vertex  ( int index ) const; // Coords of vertex (zero-based).
        Vec2   StartOf ( int comp  ) const; // First coord of component (zero-based).
        Vec2   EndOf   ( int comp  ) const; // Last coord of component (zero-based).
        Vec2   PointAt ( double s, Vec2 *tangent = NULL ) const; // Coords at arclength "s".
        double  LengthOf( int comp  ) const; // Length of component (zero-based).
        double  AreaOf  ( int comp  ) const; // Signed area of component (zero-based).
        Vec2   Closest ( const Vec2 &                ) const;
        Vec2   Closest ( const Vec2 &, Pnode* &node  ) const;
        Vec2   Closest ( const Vec2 &, double  &param ) const;
        Pnode *CompPtr ( int ) const; // Return start of the component.
        int    Inside  ( const Vec2 & ) const; // Treats multiline as a closed polygon.

    public: // Non-const member functions.
        void      Clear(); // Delete all vertices.
        void      SetWeight( short w ) { weight = w; }
        Multiline &Reverse(); // Destructive reversal.
        Multiline &Pop( Multiline & ); // Remove the first component.
        Multiline Pop(); // Remove the first component.
        Vec2      Move( int k, const Vec2 &delta );

    public: // Static constants and methods for inquiring global statistics.
        static long  NumNodes() { return num_nodes; }
        static long  NumLines() { return num_lines; }
        static const Vec2 Jump; // Marker for new segment.
        static const Multiline Null;

    private: // Methods used only by the Multiline class.
        void   Free( Pnode *from, Pnode *to );
        void   RecomputeBox();
        Pnode *NewNode( const Vec2 & ); // Allocate new link node.
        Vec2   Closest( const Vec2 &, Pnode* &node, double &param ) const;

    private: // Data associated with each Multiline instance.
        Box2d  box;    // Bounding box of entire Multiline.
        Pnode *first;  // First vertex of the the first component.
        Pnode *penult; // Next-to-last vertex of last component.
        Pnode *last;   // Last vertex of the last component.
        double  length; // Combined length of components (excluding jumps).
        double  seglen; // The length of the current component.
        int    points; // Total number of vertices.
        short  weight; // Line weight.
        short  comps;  // Number of connected components.
        short  jump;   // Indicates start of a new component.
        short  dirty;  // Indicates that bounding box may be out of date.
    };

extern Vec2 Interpolate( 
    const Pnode *, 
    const Pnode *, double 
    );

extern Multiline operator*( 
    const Trans2    &,
    const Multiline & 
    );

extern Multiline *Explode( 
    const Multiline & 
    );

extern Multiline Dashed( 
    const Multiline &, 
    double on, 
    double off, 
    double shift 
    );

extern Multiline CopyComponent( 
    const Multiline &, 
    int comp  // Zero-based component index.
    ); 

extern Multiline CopyPointsSince( 
    const Multiline &lines, 
    long start 
    );

extern Multiline Truncate( 
    const Multiline &, 
    double len 
    );

extern void Split( 
    const Multiline &, // The multiline to split.
    double s,           // The arclength parameter to split at.
    Multiline &A,      // The portion from 0 to s.
    Multiline &B       // The portion from s to the end.
    );

extern Multiline Partition( 
    const Multiline &, 
    int n, 
    double split[] 
    );

extern double GapLength( 
    const Multiline & 
    );

extern double Distance(
    const Multiline &A,
    const Multiline &B,
    double &param_A,
    double &param_B
    );

extern double SignedArea( 
    const Multiline &pgon 
    );

extern Multiline SpliceAtNearestPoint( 
    const Multiline &A, 
    const Multiline &B 
    );

extern Multiline Resample( 
    const Multiline &P, 
    double eps 
    );

typedef Multiline *MultilinePtr;

extern ostream &operator<<( ostream &, const Multiline &M );
extern istream &operator>>( istream &,       Multiline &M );

#endif

